import numpy as np
import healpy as hp
import matplotlib.pyplot as plt

class Template:
    def __init__(self, shifts_template, nside, sigma_of_shift, mean_offset=0., mask=None):
        """ Contains information about spatial shifts to the center or width of a dndz """
        self.map = shifts_template + mean_offset
        self.npix = len(shifts_template)
        self.nside = nside
        self.sigma_of_shift = sigma_of_shift
        self.mask = mask

    def show(self, temp_plot_min=None, temp_plot_max=None, title=r'$z_{\mathrm{shift}}$'):
        """ Mollview visualization of the shifts template """
        hp.mollview(self.mask * self.map, min=temp_plot_min, max=temp_plot_max, cmap='PuOr', title=title)
        plt.show()

class MockTemplate(Template):
    def __init__(self, sigma_of_shift, nside, alpha, lmax_pert, mean_offset=0., mask=None):
        """ Generate a pixelized map where each pixel value corresponds to a random shift of the
        either the center or the width of the dndz, which is assumed to be Gaussian. The map has
        a standard deviation set by sigma_of_shift, and some underlying power-law power spectrum
        for the shifts with spectral index alpha, truncated at lmax_pert
            - Inputs:
                * sigma_of_shift = float. Standard deviation of the shifts across the template
                * nside = int. Defines the pixelization
                * alpha = float. Power law index for the perturbation Cls
                * lmax_pert = int. lmax of the Cls
                * (optional) mean_offset = float. Mean value about which template must fluctuate
                    (useful for template of interloper fraction, otherwise set to zero by default)
            * (optional) mask = np.array of floats defining observed footprint
        """
        if mask is None:
            mask = np.ones_like(hp.nside2npix(nside))
        self.lmax_pert = lmax_pert
        self.alpha = alpha

        pert_cls = np.arange(lmax_pert, dtype=float)**alpha
        pert_cls[0]=0
        # Convert to a map
        sim_map = hp.synfast(pert_cls, nside)
        # Normalize Cls to give desired variance
        norm_factor_for_alms = np.nan_to_num(sigma_of_shift**2 / np.var(sim_map, ddof=1))**0.5
        shifts_template = norm_factor_for_alms*sim_map
        shifts_template *= mask
        super().__init__(shifts_template, nside, sigma_of_shift, mean_offset, mask)
        
        
class CustomTemplate(Template):
    def __init__(self, shifts_template, mean_offset=0., mask=None):
        """ User provides a template of shifts to either the center or the width of a Gaussian dndz
        - Inputs:
            * shifts_template = 1D np.array of floats. A healpy map of shifts
            * (optional) mean_offset = float. Mean value about which template must fluctuate
                    (useful for template of interloper fraction, otherwise set to zero by default)
            * (optional) mask = np.array of floats defining observed footprint
        """
        npix = len(shifts_template)
        nside = hp.npix2nside(npix)
        sigma_of_shift = np.std(shifts_template, ddof=1)
        if mask is None:
            mask = np.ones_like(npix)
        super().__init__(shifts_template, nside, sigma_of_shift, mean_offset, mask)



import numpy as np
from scipy import interpolate
from scipy.interpolate import interp1d
from py3nj import wigner3j
from scipy.integrate import quadrature
import multiprocessing
from functools import partial

#
# First, the unbiased/isotropic case
#

def unbiased_term(exp, ells, num_processes=1, miniter=1000, maxiter=2000, tol=1e-12):
    """ Calculate the unbiased contribution to the angular clustering power spectrum in the Limber approximation
        - Inputs:
            * exp = an instance of the experiment class
            * ells = np array of ints. The ells at which to evaluate this bias
            * num_processes (optional) = int. Number of processes to use. If >1, use multiprocessing
            * miniter (optional) = int. Minimum number of iterations for quadrature.
            * maxiter (optional) = int. Maximum number of iterations for quadrature.
            * tol (optional) = int. Error tolerance before breaking numerical integration.
    """
    if num_processes > 1:
        # Use multiprocessing to speed up calculation
        pool = multiprocessing.Pool(num_processes)
        print('Running in parallel with {} processes'.format(num_processes))
        # Helper function (pool.map can only take one, iterable input)
        func = partial(unbiased_term_at_l, exp, miniter, maxiter, tol)
        clgg_unbiased = np.array(pool.map(func, ells))
        pool.close()
    else:
        clgg_unbiased = np.zeros_like(ells, dtype=float)
        for i, l in enumerate(ells):
            clgg_unbiased[i] = unbiased_term_at_l(exp, miniter, maxiter, tol, l)
    return clgg_unbiased

def unbiased_term_at_l(exp, miniter, maxiter, tol, l):
    """ Calculate the unbiased contribution to the angular clustering power spectrum in the Limber approximation
        at a specific l.
        - Inputs:
            * exp = an instance of the experiment class
            * miniter (optional) = int. Minimum number of iterations for quadrature.
            * maxiter (optional) = int. Maximum number of iterations for quadrature.
            * tol (optional) = int. Error tolerance before breaking numerical integration.
            * l = int. The multipole of \Delta C_l
    """
    print('Working on l={}'.format(l))
    Pk_interp_1D = interp1d(exp.grid.chi_array, np.diagonal(np.flipud(exp.Pk_interp(exp.grid.chi_array,
                                                                               (l + 0.5) / exp.grid.chi_array))))
    result, error = quadrature(integrand_unbiased_auto_term, exp.grid.chi_min_int, exp.grid.chi_max_int,
                               args=(exp.kernel, Pk_interp_1D), miniter=miniter, maxiter=maxiter, tol=tol)
    return result

def integrand_unbiased_auto_term(chi, kernel, Pk_interp_1D):
    """
    Integrand for the unbiased Cl auto spectrum in the Limber approximation
    """
    return Pk_interp_1D(chi) * kernel(chi) / chi** 2


#
# On to the additive bias to galaxy clustering
#

def additive_bias(exp, ells, num_processes=1, miniter=1000, maxiter=2000, tol=1e-12):
    """ Calculate the mode-coupling bias to the galaxy clustering power spectrum
        - Inputs:
            * exp = an instance of the experiment class
            * ells = np array of ints. The ells at which to evaluate this bias
            * num_processes (optional) = int. Number of processes to use. If >1, use multirocessing
            * miniter (optional) = int. Minimum number of iterations for quadrature.
            * maxiter (optional) = int. Maximum number of iterations for quadrature.
            * tol (optional) = int. Error tolerance before breaking numerical integration.
    """
    if num_processes > 1:
        # Use multiprocessing to speed up calculation
        print('Running in parallel with {} processes'.format(num_processes))
        pool = multiprocessing.Pool(num_processes)
        # Helper function (pool.map can only take one, iterable input)
        func = partial(additive_bias_at_l, exp, miniter, maxiter, tol)
        additive_bias = np.array(pool.map(func, ells))
        pool.close()
    else:
        additive_bias = np.zeros_like(ells, dtype=float)
        for i, l in enumerate(ells):
            additive_bias[i] = additive_bias_at_l(exp, miniter, maxiter, tol, l)
    return additive_bias

def additive_bias_at_l(exp, miniter, maxiter, tol, l):
    """ Calculate the additive bias to the galaxy clustering power spectrum
        at a specific l.
        - Inputs:
            * exp = an instance of the experiment class
            * miniter (optional) = int. Minimum number of iterations for quadrature.
            * maxiter (optional) = int. Maximum number of iterations for quadrature.
            * tol (optional) = int. Error tolerance before breaking numerical integration.
            * l = int. The multipole of \Delta C_l
    """
    print('Working on l={}'.format(l))
    try:
        Clchi1chi2_interp = interpolate.RegularGridInterpolator((exp.grid.chi_array, exp.grid.chi_array),
                                                                exp.Cl_deltap_of_chi1_chi2[l, :, :],
                                                                method='linear', bounds_error=True, fill_value=0)

        result, error = quadrature(integrand_additive_term, exp.grid.chi_min_int,
                                   exp.grid.chi_max_int, args=(Clchi1chi2_interp, exp.grid.chi_min_int, exp.grid.chi_max_int),
                                   miniter=miniter, maxiter=maxiter, tol=tol)
    except IndexError as e:
        print(f"{e}")
        print('Setting the result to 0 because ClDeltaphi ought to be 0 at this l')
        result = 0
    return result

def integrand_additive_term(chi1, Clchi1chi2_interp, chi_min_int, chi_max_int):
    """
    Integrand for the additive bias term, in the Limber approximation
    """
    outer_integral, error = quadrature(integrand_nested_additive_term, chi_min_int,
                                       chi_max_int, args=(chi1, Clchi1chi2_interp), miniter=3,
                                       maxiter=5, tol=1e-20)
    return outer_integral

def integrand_nested_additive_term(chi2, chi1, Clchi1chi2_interp):
    chi1 = chi1[0]
    chi2 = chi2[0]
    return Clchi1chi2_interp(np.array((chi1, chi2)))[0]

#
# And, finally, the mode-coupling term
#

def mode_coupling_bias(exp, ells, lprime_max='none', num_processes=1, miniter=1000, maxiter=2000, tol=1e-12,
                       mode='full'):
    """ Calculate the mode-coupling bias to the angular clustering power spectrum in the Limber approximation
        - Inputs:
            * exp = an instance of the experiment class
            * ells = np array of ints. The ells at which to evaluate this bias
            * lprime_max (optional) = int. Value of l above which we ignore the anisotropy in C_l^{\Delta \phi}
            * num_processes (optional) = int. Number of processes to use. If >1, use multirocessing
            * miniter (optional) = int. Minimum number of iterations for quadrature.
            * maxiter (optional) = int. Maximum number of iterations for quadrature.
            * tol (optional) = int. Error tolerance before breaking numerical integration.
            * mode (optional) = str. If 'full' do the full calc (still assumes Limber).
                                    If 'analytic_via_variance' use analytic kernel \mathrm{Var}[\phi](\chi)
                                    If 'analytic_toy_model' use analytic kernel [f(\chi)\bar{\phi}(\chi)]^2
    """
    if mode == 'full':
        # Do the full calculation, using no analytic approximations
        integral_at_l = mode_coupling_bias_at_l
        if lprime_max == 'none':
            lprime_max = exp.Cl_deltap_of_chi1_chi2.shape[0] - 1
    elif mode == 'analytic_via_variance':
        # Use the analytic approximation where the Limber kernel is \mathrm{Var}[\phi](\chi)
        exp.mc_kernel = exp.analytic_proj_kernel
        integral_at_l = analytic_mode_coupling_bias_at_l
    elif mode == 'analytic_toy_model':
        # Use the analytic approximation where the Limber kernel is [f(\chi)\bar{\phi}(\chi)]^2
        exp.mc_kernel = exp.analytic_kernel_toy_model
        integral_at_l = analytic_mode_coupling_bias_at_l
    print('Using mode {}'.format(mode))

    if num_processes > 1:
        # Use multiprocessing to speed up calculation
        print('Running in parallel with {} processes'.format(num_processes))
        pool = multiprocessing.Pool(num_processes)
        # Helper function (pool.map can only take one, iterable input)
        func = partial(integral_at_l, exp, lprime_max, miniter, maxiter, tol)
        conv_bias = np.array(pool.map(func, ells))
        pool.close()
    else:
        conv_bias = np.zeros_like(ells, dtype=float)
        for i, l in enumerate(ells):
            conv_bias[i] = integral_at_l(exp, lprime_max, miniter, maxiter, tol, l)
    return conv_bias

def mode_coupling_bias_at_l(exp, lprime_max, miniter, maxiter, tol, l):
    """ Calculate the mode-coupling bias to the angular clustering power spectrum in the Limber approximation,
        at a specific l.
        - Inputs:
            * exp = an instance of the experiment class
            * lprime_max = int. Value of l above which we ignore the anisotropy in C_l^{\Delta \phi}
            * miniter (optional) = int. Minimum number of iterations for quadrature.
            * maxiter (optional) = int. Maximum number of iterations for quadrature.
            * tol (optional) = int. Error tolerance before breaking numerical integration.
            * l = int. The multipole of \Delta C_l
    """
    #TODO: Implement beyond-Limber
    print('Working on l={}'.format(l))
    integrand = np.zeros_like(exp.grid.chi_array)
    cldp_interp = exp.cldp_interp(exp.grid.chi_array)
    for L in range(l + lprime_max + 1):
        abslmL = np.abs(l - L)
        lpL = l + L
        if abslmL <= lprime_max:
            # Only loop if you will probe scales below cut
            Pk_interp = np.diagonal(np.flipud(exp.Pk_interp(exp.grid.chi_array, (L + 0.5) / exp.grid.chi_array)))
            for lprime in np.arange(abslmL, min(lprime_max, lpL) + 1, 1):
                if (l + lprime + L) % 2 == 0:
                    w3 = wigner3j(2 * l, 2 * L, 2 * lprime, 0, 0, 0)
                    prefactor = w3 ** 2 * (2 * lprime + 1) * (2 * L + 1) / (4 * np.pi)
                    integrand += prefactor / exp.grid.chi_array ** 2 * Pk_interp * cldp_interp[lprime]
    f = interp1d(exp.grid.chi_array, integrand)
    result, error = quadrature(f, exp.grid.chi_min_int, exp.grid.chi_max_int, miniter=miniter, maxiter=maxiter, tol=tol)
    return result

def analytic_mode_coupling_bias_at_l(exp, dummy, miniter, maxiter, tol, l):
    """ Calculate the mode-coupling bias to the angular clustering power spectrum in the Limber approximation,
        at a specific l (l \gg 1 for this approach to be valid), using the variance of \Delta Phi in each chi slice
        - Inputs:
            * exp = an instance of the experiment class
            * dummy = whatever. We don't actually need this, but we have them in order to match the code structure
            * miniter (optional) = int. Minimum number of iterations for quadrature.
            * maxiter (optional) = int. Maximum number of iterations for quadrature.
            * tol (optional) = int. Error tolerance before breaking numerical integration.
            * l = int. The multipole of \Delta C_l
    """
    print('Working on l={}'.format(l))
    # Interpolate at the scales required by Limber
    Pk_interp_1D = interp1d(exp.grid.chi_array, np.diagonal(np.flipud(exp.Pk_interp(exp.grid.chi_array,
                                                                               (l + 0.5) / exp.grid.chi_array))))
    result, error = quadrature(limber_integral, exp.grid.chi_min_int,
                               exp.grid.chi_max_int, args=(Pk_interp_1D, exp.mc_kernel),
                               miniter=miniter, maxiter=maxiter, tol=tol)
    return result

def limber_integral(chi, Pk_interp_1D, kernel):
    return kernel(chi) * Pk_interp_1D(chi)









import numpy as np
import healpy as hp
from scipy import interpolate
import matplotlib.pyplot as plt
from scipy.ndimage.filters import gaussian_filter
import pickle
from astropy.cosmology import Planck18
from cardiac import utils
from cardiac import tracer_spectra
from cardiac import bias_integrals

class Spec:
    def __init__(self, field1=None, field2=None, load=False, save=False, filename=None, get_cls=False,
                 gbias_mode='linear'):
        """ General class defining a spectrum between two fields, including infrastruture to compute all contributions
            - Inputs:
                * field1 (optional) = Instance of fields.Field class. Needed when not loading from file
                * field2 (optional) = Instance of fields.Field class. Defaults to field1
                * load (optional) = Bool. Whether to load object from file. If True, filename must be provided
                * save (optional) = Bool. Whether to save object to file. If True, filename must be provided
                * filename (optional) = str. Path to pickled object to write to / load from
                * get_cls (optional) = Bool. Whether to calculate all contributions on this first pass
                * gbias_mode (optional) = 'linear' or 'anzu'. Galaxy bias prescription
        """
        assert (field1 is not None or load!=False), "Must either initialize new object or load one from file!"
        if (load!=False) or (save!=False):
            assert (filename is not None), "Must provide a filename when saving or loading"
        if field2 is None:
            field2 = field1
        assert(field1 == field2), "Fields are incompatible given their grids"

        if load:
            with open(filename + '.pkl', 'rb') as input:
                self.__dict__ = pickle.load(input)
        else:
            self.field1 = field1
            self.field2 = field2
            self.label_dict = {'GalDelta': r'\phi', 'GalShear': r'g'}
            self.grid = field1.grid

            # Measure sky fraction covered by footprint. Need it to de-bias measurement of Cls / covariance of phi
            # TODO: generalize to case where footprint of field1 is different from that of field2
            assert (np.array_equal(self.field1.mask,
                                   self.field2.mask)), "Implement case where footprint of field1 is different from that of field2"
            self.fsky = np.mean(self.field1.mask)

            # Get the (co)variance btw delta_p's in each slice
            # From this we can directly calculate mode-coupling biases at large l
            covmat = np.cov(field1.delta_p_maps, field2.delta_p_maps, ddof=1, rowvar=False)
            self.cov_at_chi = np.diagonal(covmat, offset=field1.grid.n_samples_of_chi) / self.fsky

            # The fiducial projection kernels
            self.kernel = interpolate.interp1d(self.grid.chi_array, field1.p_fid_array * field2.p_fid_array)

            # The kernels in Limber integral when approximating the mode-coupling bias in the limit l>>1
            self.analytic_proj_kernel = interpolate.interp1d(self.grid.chi_array, self.cov_at_chi/self.grid.chi_array**2)

        if get_cls:
            self.get_contributions(gbias_mode=gbias_mode)
        if save:
            self.save_properties(filename)

    def save_properties(self, output_filename='./dict_with_properties'):
        """
        Save the dictionary of key properties to file
        Inputs:
            * output_filename = str. Output filename
        """
        with open(output_filename+'.pkl', 'wb') as output:
            pickle.dump(self.__dict__, output, pickle.HIGHEST_PROTOCOL)

    def get_Cldp1dp2(self):
        """ Calculate the multi-frequency angular power spectrum of our perturbations (MAPS), a key ingredient
            of the bias calculations
        """
        self.lmax = hp.Alm.getlmax(self.field1.delta_p_lm_of_chi.shape[0])
        self.Cl_deltap_of_chi1_chi2 = np.zeros((self.lmax + 1, self.field1.grid.n_samples_of_chi,
                                                self.field2.grid.n_samples_of_chi))
        # ToDo: JIT this?
        for i in range(self.field1.delta_p_lm_of_chi.shape[0]):
            l, m = hp.Alm.getlm(self.lmax, i)  # Get the l corresponding to each value of m
            if m != 0:
                # Healpix indexes only m (not -m, since Ylm=Yl-m for a real field), so correct the sum for this
                factor = 2
            else:
                factor = 1
            # Get angular PS and deconvolve pixel window function for all possible combinations of chi1 and chi2
            # We correct for partial-sky coverage using a simple fsky correction.
            # TODO: implement Namaster or similar
            self.Cl_deltap_of_chi1_chi2[l, :, :] += factor * np.outer(self.field1.delta_p_lm_of_chi[i, :],
                                                                      np.conj(
                                                                          self.field2.delta_p_lm_of_chi[i, :])).real / (
                                                            2 * l + 1) / self.fsky

        # We'll need the interpolated version of this
        self.cldp_interp = interpolate.interp1d(self.field1.grid.chi_array,
                                    np.diagonal(self.Cl_deltap_of_chi1_chi2, axis1=1, axis2=2), axis=-1)

    def get_contributions(self, ells=None, num_processes=1, miniter=1000, maxiter=5000, tol=1e-11, lprime_max='none',
                          gbias_mode='linear'):
        """
        Helper function to calculate all the contributions to angular clustering -- isotropic and anisotropic
        - Inputs:
        * ells (optional) = np.array of ints. Multipoles at which to evauate the contributions
        * num_processes (optional) = int. Number of processes to use. If >1, use multirocessing
        * miniter (optional) = int. Minimum number of iterations for quadrature.
        * maxiter (optional) = int. Maximum number of iterations for quadrature.
        * tol (optional) = int. Error tolerance before breaking numerical integration.
        * lprime_max (optional) = int. Value of l above which we ignore the anisotropy in C_l^{\Delta \phi}
        * gbias_mode (optional) = 'linear' or 'anzu'. Galaxy bias prescription
        """
        if not hasattr(self, 'Cl_deltap_of_chi1_chi2'):
            self.get_Cldp1dp2()
        if not hasattr(self, 'Pk'):
            self.get_3D_spectrum(gbias_mode=gbias_mode)

        if ells is None:
            # The ells where we want to evaluate the spectra
            self.ells = np.logspace(np.log10(50), np.log10(1500), 48, dtype=int)
        self.unbiased_cls = bias_integrals.unbiased_term(self, self.ells, num_processes=num_processes, miniter=miniter,
                                      maxiter=maxiter, tol=tol)

        # Compute full mode-coupling bias
        self.conv_bias = bias_integrals.mode_coupling_bias(self, self.ells, lprime_max=lprime_max,
                                                  num_processes=num_processes,
                                                  miniter=miniter, maxiter=maxiter, tol=tol)

        # And the analytic approximation, which is very good on small scales
        self.analytic_mcbias_via_variance = bias_integrals.mode_coupling_bias(self, self.ells,
                                                                     num_processes=num_processes,
                                                                     miniter=miniter, maxiter=maxiter, tol=tol,
                                                                     mode='analytic_via_variance')

        if self.field1.__class__.__name__=='GalDelta' and self.field2.__class__.__name__=='GalDelta':
            # Compute additive bias
            self.additive_bias = bias_integrals.additive_bias(self, self.ells, num_processes=num_processes,
                                                     miniter=miniter, maxiter=maxiter, tol=tol)

    def get_3D_spectrum(self, gbias_mode='linear'):
        """ Helper function to calculate the 3D power spectra of the tracers. Uses functions in tracer_spectra.py,
            which in turn call CAMB/anzu. These can be easily replaced with whatever the user prefers
            Input:
                * gbias_mode (optional) = 'linear' or 'anzu'. Galaxy bias prescription
        """
        higher_z_mean = max(self.field1.z_mean, self.field1.z_mean)  # Central redshift of the furthest away tracer
        lower_z_mean = min(self.field1.z_mean, self.field1.z_mean)  # Central redshift of the furthest away tracer
        higher_sigma = max(self.field1.sigma, self.field1.sigma)  # The higher of the widths
        if self.field1.__class__.__name__=='GalDelta' and self.field2.__class__.__name__=='GalDelta':
            # Get galaxy power spectrum at redshifts near the center of the dN/dz
            # Evaluate predictions at the Planck 18 cosmology and redshifts within 5sigma of the dndzmean
            zs_sampled = np.linspace(max(lower_z_mean - 5 * higher_sigma, 0.01), higher_z_mean + 5 * higher_sigma, 30) 
            chis_sampled = Planck18.comoving_distance(zs_sampled).value

            if gbias_mode=='linear':
                assert(isinstance(self.field1.bvec, float) and isinstance(self.field2.bvec, float)), "Galaxy bias must be a float" \
                                                                               " when using linear bias prescription"
            else:
                assert (len(self.field1.bvec) > 1 and len(
                    self.field2.bvec) > 1), "Please provide all Lagrangian bias parameters required by anzu"

            k, self.Pk = tracer_spectra.get_galaxy_ps(self.field1.bvec, zs_sampled, g2_bias=self.field2.bvec,
                                                      gbias_mode=gbias_mode)
        elif all(x in [self.field1.__class__.__name__, self.field2.__class__.__name__] for x in ['GalDelta', 'GalShear']):
            # Get galaxy-cross-matter power spectrum at a few redshifts around the lens galaxy sample
            if self.field1.__class__.__name__ == 'GalDelta':
                gal_z_mean = self.field1.z_mean
                gbias = self.field1.bvec
            else:
                gal_z_mean = self.field2.z_mean
                gbias = self.field2.bvec
            zs_sampled = np.linspace(max(gal_z_mean - 7 * higher_sigma, 0.01), gal_z_mean + 7 * higher_sigma, 30)
            chis_sampled = Planck18.comoving_distance(zs_sampled).value

            if gbias_mode=='linear':
                assert(isinstance(gbias, float)), "Galaxy bias must be a float when using linear bias prescription"
            else:
                assert (len(gbias) > 1), "Please provide all Lagrangian bias parameters required by anzu"

            k, self.Pk = tracer_spectra.get_galaxy_ps(gbias, zs_sampled, g2_bias=None, gbias_mode=gbias_mode)
        elif self.field1.__class__.__name__ == 'GalShear' and self.field2.__class__.__name__ == 'GalShear':
            # Get the non-linear matter power spectrum at redshifts btw 0 and the source galaxies
            zs_sampled = np.linspace(0, higher_z_mean + 5 * higher_sigma)
            chis_sampled = Planck18.comoving_distance(zs_sampled).value
            k, pk_nonlin = tracer_spectra.get_matter_ps(zs_sampled)
            self.Pk = np.swapaxes(pk_nonlin, 0, 1)

        # Interpolate
        # Note: scipy.interp2d is deprecated, but it is MUCH faster than the new alternatives...
        self.Pk_interp = interpolate.interp2d(chis_sampled, k, self.Pk,
                                                kind='linear', bounds_error=False, fill_value=0)

    def plot_DeltaP_covariance(self):
        '''
        Plot the perturbation covariance as a function of distance
        '''
        plt.axhline(0, color='gray', lw=0.5)
        plt.axvline(self.field1.chi_mean_fid, color='k', ls='--')
        plt.plot(self.grid.chi_array, self.cov_at_chi)
        plt.ylabel(r'Cov[$\Delta '+self.label_dict[self.field1.__class__.__name__]+'(\chi), \Delta '+
                   self.label_dict[self.field2.__class__.__name__]+'(\chi)$] [Mpc$^{-2}$]')
        plt.xlabel(r'$\chi$ [Mpc]')
        plt.show()

    def plot_Cl_DeltaP_of_chi(self):
        '''
        Plot C_l^{\Delta \phi}(\chi,\chi) vs \mathrm{log}_{10}\,\chi for various l's
        '''
        for l_to_plot in np.linspace(10, self.lmax, 8, dtype=int):
            plt.plot(self.grid.chi_array, np.diagonal(self.Cl_deltap_of_chi1_chi2, axis1=1, axis2=2)[l_to_plot, :],
                     label=r'$l={}$'.format(l_to_plot))
        plt.axvline(self.field1.chi_mean_fid, ls='--', color='k', label=r'$r=\chi(z_{\mathrm{mean}})$')
        plt.xlabel(r'$\mathrm{log}_{10}\,\chi$')
        plt.ylabel(r'$C_l^{\Delta '+self.label_dict[self.field1.__class__.__name__]+'}(\chi,\chi)$')
        plt.legend()
        plt.xlim([1, self.field1.chi_mean_fid + 10*self.field1.chi_sigma_fid])
        plt.show()

    def plot_Cl_DeltaP_of_chi_2D(self, xmin=None, xmax=None, min_log_range=-20, max_log_range=-9):
        '''
        Plot log C_l^{\Delta \phi}(\chi,\chi) against l and  \mathrm{log}_{10}\,\chi
        '''
        if xmin is None:
            xmin = self.field1.chi_mean_fid - 7*self.field1.chi_sigma_fid
        if xmax is None:
            xmax = self.field1.chi_mean_fid + 7*self.field1.chi_sigma_fid

        X, Y = np.meshgrid(np.arange(len(self.grid.chi_array)), np.arange(self.lmax + 1))
        Z = np.log10(np.diagonal(self.Cl_deltap_of_chi1_chi2, axis1=1, axis2=2))
        # Set the log of 0 to a tiny negative number
        Z[np.diagonal(self.Cl_deltap_of_chi1_chi2, axis1=1, axis2=2) == 0] = -1e100

        contours = np.linspace(min_log_range, max_log_range, 20)

        # Smooth array with a Gaussian filter for plotting purposes
        Z_smoothed = gaussian_filter(Z, sigma=3)
        plt.contourf(X, Y, Z_smoothed, levels=contours, cmap='inferno', extend='both')

        ax = plt.gca()
        ax.axvline(np.where(self.grid.chi_array > self.field1.chi_mean_fid)[0][0], color='r', ls='--', lw=1,
                   label=r'$r=\chi(z_{\mathrm{mean}})$')
        ax.set_ylim([0, self.lmax])

        plt.xlabel(r'$\mathrm{log}_{10}\,\chi$')
        plt.ylabel(r'$l$')

        label_locs = np.round(np.linspace(xmin+50, xmax-50, 3, dtype=int), decimals=-1)
        ax.set_xticks(utils.find_closest_indices(self.grid.chi_array, label_locs))
        ax.set_xticklabels(label_locs.astype('str'))
        plt.legend()

        plt.title(r'$\mathrm{log}_{10} \, C_l^{\Delta '+self.label_dict[self.field1.__class__.__name__]+'}(\chi)$')
        plt.colorbar(location='right')
        plt.xlim([np.where(self.grid.chi_array > xmin)[0][0], np.where(self.grid.chi_array > xmax)[0][0]])
        plt.show()

    def plot_Cl_DeltaP_of_chi_chiprime_2D(self, xmin=None, xmax=None, min_log_range=-20, max_log_range=-9):
        '''
        Plot log |C_l^{\Delta \phi}(\chi,\chi')| against l and  \mathrm{log}_{10}\,\chi
        '''
        if xmin is None:
            xmin = self.field1.chi_mean_fid - 7*self.field1.chi_sigma_fid
        if xmax is None:
            xmax = self.field1.chi_mean_fid + 7*self.field1.chi_sigma_fid

        chi_idx = np.where(np.log10(self.grid.chi_array) > 3.2)[0][0]
        X, Y = np.meshgrid(np.arange(len(self.grid.chi_array)), np.arange(self.lmax + 1))

        Z_fixed_chi = np.log10(np.abs(self.Cl_deltap_of_chi1_chi2[:, :, chi_idx]))
        # Set the log of 0 to a tiny negative number
        Z_fixed_chi[np.abs(self.Cl_deltap_of_chi1_chi2[:, :, chi_idx]) == 0] = -1e100

        # Smooth array with a Gaussian filter for plotting purposes
        Z_fixed_chi_smoothed = gaussian_filter(Z_fixed_chi, sigma=3)

        contours = np.linspace(min_log_range, max_log_range, 20)
        plt.contourf(X, Y, Z_fixed_chi_smoothed, levels=contours, cmap='inferno', extend='both')

        plt.xlabel(r'$\mathrm{log}_{10}\,\chi$')
        plt.ylabel(r'$l$')

        ax = plt.gca()
        ax.axvline(np.where(self.grid.chi_array > self.field1.chi_mean_fid)[0][0], color='r', ls='--', lw=1,
                   label=r'$r=r(z_{\mathrm{mean}})$')

        ax.set_ylim([0, self.lmax])

        label_locs = np.round(np.linspace(xmin+50, xmax-50, 3, dtype=int), decimals=-1)
        ax.set_xticks(utils.find_closest_indices(self.grid.chi_array, label_locs))
        ax.set_xticklabels(label_locs.astype('str'))

        plt.title(r'$\mathrm{log}_{10} \, |C_l^{\Delta '+self.label_dict[self.field1.__class__.__name__]+'}(\chi,\chi^{*})|$')
        plt.colorbar(location='right')

        ax.axvline(chi_idx, label=r'$\chi^{*}$', ls=':', color='w')
        plt.legend()
        plt.xlim([np.where(self.grid.chi_array > xmin)[0][0], np.where(self.grid.chi_array > xmax)[0][0]])
        plt.show()

    def plot_Cl_DeltaP_vs_chi_chiprime_2D(self, l, color_range=None, xmin=None, xmax=None):
        '''
        Plot C_l^{\Delta \phi}(\chi,\chi') against \chi \chi' a fixed l
        '''
        if xmin is None:
            xmin = self.field1.chi_mean_fid - 5*self.field1.chi_sigma_fid
        if xmax is None:
            xmax = self.field1.chi_mean_fid + 5*self.field1.chi_sigma_fid            
        
        suffix = ''

        X, Y = np.meshgrid(np.arange(len(self.grid.chi_array)), np.arange(len(self.grid.chi_array)))
        Z = self.Cl_deltap_of_chi1_chi2[l, :, :]
        
        if color_range is None:
            color_range = np.max(Z)
        contours = np.linspace(-color_range, color_range, 20)

        plt.contourf(X, Y, Z, cmap='RdBu', levels=contours, extend='neither')

        ax = plt.gca()
        ax.axvline(np.where(self.grid.chi_array > self.field1.chi_mean_fid)[0][0],
                   color='k', ls=':', lw=1, label=r'$\chi=\chi_0$')
        ax.axhline(np.where(self.grid.chi_array > self.field1.chi_mean_fid)[0][0],
                   color='k', ls=':', lw=1)

        plt.xlabel(r'$\chi_1$ [Mpc]')
        plt.ylabel(r'$\chi_2$ [Mpc]')

        label_locs = np.round(np.linspace(xmin+50, xmax-50, 3, dtype=int), decimals=-1)
        ax.set_xticks(utils.find_closest_indices(self.grid.chi_array, label_locs))
        ax.set_yticks(utils.find_closest_indices(self.grid.chi_array, label_locs))

        ax.set_xticklabels(label_locs.astype('str'))
        ax.set_yticklabels(label_locs.astype('str'))
        plt.legend()

        plt.title(r'$ C_\ell^{\Delta '+self.label_dict[self.field1.__class__.__name__]+
                  ' \Delta '+self.label_dict[self.field2.__class__.__name__]+'}(\chi_1, \chi_2)$ for $\ell=$' + str(l))
        cbar = plt.colorbar(location='right')
        cbar_labels = np.linspace(np.min(contours), np.max(contours), 5, dtype=float)
        cbar.set_ticks(cbar_labels)
        cbar.formatter.set_powerlimits((0, 0))
        cbar.formatter.set_useMathText(True)

        plt.xlim([np.where(self.grid.chi_array > xmin)[0][0], np.where(self.grid.chi_array > xmax)[0][0]])
        plt.ylim([np.where(self.grid.chi_array > xmin)[0][0], np.where(self.grid.chi_array > xmax)[0][0]])
        plt.show()



import numpy as np
import healpy as hp
from scipy.interpolate import interp1d
from scipy.integrate import quadrature
from astropy.cosmology import Planck18, z_at_value
from astropy.constants import c
import astropy.units as u
from cardiac import utils

class grid:
    def __init__(self, nside, n_samples_of_chi, z_min_int=0.005, z_max_int = 3.):
        """ Defines the numerical hyperparameters of our calculation
            - Inputs:sigma_chishift
                * nside = int, power of 2. Healpix nside defining the pixelization. Must match shifts template
                * z_min_int (optional) = float. Minimum of range for the integrals over chi
                * z_max_int (optional) = float. z_max_int
                * n_samples_of_chi = int. A power of 2, # of samples in comoving distance
        """
        self.nside = nside
        self.npix = hp.nside2npix(nside)
        self.n_samples_of_chi = n_samples_of_chi
        self.chi_min_int = Planck18.comoving_distance(z_min_int).value
        self.chi_max_int = Planck18.comoving_distance(z_max_int).value
        self.chi_array = np.linspace(self.chi_min_int, self.chi_max_int + 100, n_samples_of_chi)
        # Get the redshift corresponding to these values of chi in the Planck18 cosmology
        self.z_array = np.zeros((self.npix, n_samples_of_chi))
        for i, chi in enumerate(self.chi_array):
            self.z_array[:, i] = z_at_value(Planck18.comoving_distance, chi * u.Mpc)

    def hashdict(self):
        return {'n_samples_of_chi': self.n_samples_of_chi,
                'nside': self.nside, 'chi_min_int': self.chi_min_int, 'chi_max_int': self.chi_max_int}

    def __eq__(self, other):
        return self.compatible(other)

    def compatible(self, other):
        return ((self.n_samples_of_chi == other.n_samples_of_chi) and
                (self.nside == other.nside)  and
                (self.chi_min_int == other.chi_min_int) and
                (self.chi_max_int == other.chi_max_int))

class Field:
    def __init__(self, grid, p_pert_array, p_fid_array, mask):
        """ General class defining observed fields subject to projection anisotropy. Computes analytic projection kernel
            and alms of each redshift slice
            - Inputs:
                * grid = Instance of grid class containing numerical hyperparams
                * p_pert_array = np.array of size (grid.npix, grid.n_samples_of_chi). Contains perturbation to phi/g/etc
                * p_fid_array = np.array of size (grid.n_samples_of_chi). Fiducial phi/g/etc
                * mask = np.array of float values between 0 and 1 defining observed footprint
        """
        self.delta_p_maps = mask[..., None] * (p_pert_array - p_fid_array)
        self.p_fid_array = p_fid_array
        self.grid = grid
        self.mask = mask

        # Take the spherical harmonic transform of each chi slice.
        # Conveniently, we can take complex SHT so array sizes reduce by x2
        self.delta_p_lm_of_chi = np.zeros((hp.Alm.getsize(3 * grid.nside - 1), grid.n_samples_of_chi), dtype=complex)
        for i in range(grid.n_samples_of_chi):
            self.delta_p_lm_of_chi[:, i] = hp.map2alm(self.delta_p_maps[:, i])

    # ToDo: Maybe 'equality' isn't the clearest way of checking this
    def __eq__(self, other):
        return self.compatible(other)

    def compatible(self, other):
        return self.grid == other.grid

class GalDelta(Field):
    def __init__(self, grid, sigma, z_mean, bvec, template_zmean_shifts=None, template_width_shifts=None,
                 template_interloper_frac=None, interloper_sigma=None, interloper_z_mean=None, get_delta_p=True):
        """ Observed galaxy clustering field subject to anisotropy in its (Gaussian) dN/dz
            - Inputs:
                * grid = Instance of grid class containing numerical hyperparams
                * sigma = float. Standard deviation of the fiducial dN/dz
                * z_mean = float. Central redshift of the fiducial dN/dz
                * bvec = list containing [b1,    b2,    bs2,   bnabla2, SN] to be fed to Anzu to obtain Pgg
                * template_zmean_shifts = instance of templates.Template with shifts in the mean redshift of the dN/dz
                * template_width_shifts = instance of templates.Template with shifts in the width of the dN/dz
                * template_interloper_frac = instance of templates.Template with fraction of main Gaussian that's in
                                             the secondary one
                * interloper_sigma = float. Standard deviation of the fiducial interloper dN/dz
                * interloper_z_mean = float. Central redshift of the fiducial interloper dN/dz
                * get_delta_p (optional) = Bool. If False, use as helper function in cosmic shear calculation
        """
        if template_zmean_shifts is not None:
            assert (grid.npix==len(template_zmean_shifts.map)), "grid does not match nside of z_mean shift template"
            self.mask = template_zmean_shifts.mask
        if template_width_shifts is not None:
            assert (grid.npix==len(template_width_shifts.map)), "grid does not match nside of width shift template"
            self.mask = template_width_shifts.mask
        if template_interloper_frac is not None:
            assert (interloper_sigma is not None and interloper_z_mean is not None), "Provide interloper dN/dz!"
            assert (grid.npix==len(template_interloper_frac.map)), "grid does not match nside of interloped f. template"
            self.mask = template_interloper_frac.mask

        self.template_zmean_shifts = template_zmean_shifts
        self.template_width_shifts = template_width_shifts
        self.template_interloper_frac = template_interloper_frac
        self.sigma = sigma
        self.z_mean = z_mean
        self.interloper_sigma = interloper_sigma
        self.interloper_z_mean = interloper_z_mean
        self.bvec = bvec # ToDo: Allow tracers to have different bias
        # The user input is in redshift units because this is more intuitive. However, we will define our dndzs to be
        # Gaussian in comoving distance. So next, we convert to chi
        self.chi_mean_fid = Planck18.comoving_distance(z_mean).value
        # Width of the fiducial distribution
        self.chi_sigma_fid = Planck18.comoving_distance(z_mean + sigma).value - Planck18.comoving_distance(z_mean).value

        # Convert template of z-shifts to chi-shifts
        if template_zmean_shifts is None:
            chimean_shifts_array = np.zeros(grid.npix, dtype=float)
        else:
            chimean_shifts_array = Planck18.comoving_distance(z_mean + template_zmean_shifts.map).value \
                                   - Planck18.comoving_distance(z_mean).value
        if template_width_shifts is None:
            width_shifts_array = np.zeros(grid.npix, dtype=float)
        else:
            width_shifts_array = Planck18.comoving_distance(z_mean + template_width_shifts.map).value \
                                 - Planck18.comoving_distance(z_mean).value

        if template_interloper_frac is not None:
            # If studying interlopers, ignore variations in main Gaussian
            print('Studying interlopers ONLY and ignoring any other anisotropy')
            # ToDo: allow for both interlopers and anisotropy in bulk of dndz
            self.interloper_chi_mean_fid = Planck18.comoving_distance(interloper_z_mean).value
            self.interloper_chi_sigma_fid = Planck18.comoving_distance(interloper_z_mean + interloper_sigma).value \
                                            - Planck18.comoving_distance(interloper_z_mean).value
            # TODO: Implement non-gaussian dndz
            dndz_main = (1 / (  (self.chi_sigma_fid) * np.sqrt(2 * np.pi))) * \
                             np.exp( -(grid.chi_array - self.chi_mean_fid ) ** 2 / ( 2 * (self.chi_sigma_fid) ** 2))
            dndz_interloper = (1 / ((self.interloper_chi_sigma_fid) * np.sqrt(2 * np.pi))) * \
                        np.exp(-(grid.chi_array - self.interloper_chi_mean_fid) ** 2
                               / (2 * (self.interloper_chi_sigma_fid) ** 2))
            dndz_perturbed = (1-template_interloper_frac.map[..., np.newaxis])*dndz_main \
                             + template_interloper_frac.map[..., np.newaxis]*dndz_interloper
        else:
            # In each pixel, calculate the perturbed dndz as a Gaussian in chi
            dndz_perturbed = (1 / ((self.chi_sigma_fid + width_shifts_array[..., np.newaxis]) * np.sqrt(2 * np.pi))) * np.exp(
                -(grid.chi_array - self.chi_mean_fid - chimean_shifts_array[..., np.newaxis]) ** 2 / (
                            2 * (self.chi_sigma_fid + width_shifts_array[..., np.newaxis]) ** 2))

        # Take the fiducial dndz to be the monopole of the perturbed dndz
        dndz_fid = np.mean(dndz_perturbed, axis=0, where=self.mask[..., None].astype(bool))

        # Convert dndz to selection function
        phi_perturbed_array = (Planck18.H(grid.z_array[0, :]) / c).value * dndz_perturbed
        phi_fid_array = (Planck18.H(grid.z_array[0, :]) / c).value * dndz_fid

        # Normalize the selection function so that \int_{0}^{inf} dr phi(r) = 1
        phi_norm, error = quadrature(interp1d(grid.chi_array, phi_fid_array), grid.chi_min_int, grid.chi_max_int,
                                     tol=1e-20, miniter=3000)
        phi_perturbed_array *= phi_norm ** (-1)
        phi_fid_array *= phi_norm ** (-1)

        if get_delta_p:
            # Go on to extract the alms at each chi, and so on
            super().__init__(grid, phi_perturbed_array, phi_fid_array, self.mask)
        else:
            # Alternatively, when doing the cosmic shear calculation, we only need these two
            self.phi_perturbed_array = phi_perturbed_array
            self.phi_fid_array = phi_fid_array

class GalShear(Field):
    def __init__(self, grid, sigma, z_mean, template_zmean_shifts=None, template_width_shifts=None):
        """ Observed cosmic shear field subject to anisotropy in the (Gaussian) dN/dz of the source galaxies
            - Inputs:
                * grid = Instance of grid class containing numerical hyperparams
                * sigma = float. Standard deviation of the fiducial dN/dz
                * z_mean = float. Central redshift of the fiducial dN/dz
                * template_zmean_shifts = instance of templates.Template with shifts in the mean redshift of the dN/dz
                * template_width_shifts = instance of templates.Template with shifts in the width of the dN/dz
        """
        self.sigma = sigma
        self.z_mean = z_mean
        self.template_zmean_shifts = template_zmean_shifts
        self.template_width_shifts = template_width_shifts
        if template_zmean_shifts is not None:
            self.mask = template_zmean_shifts.mask
        else:
            self.mask = template_width_shifts.mask

        g_d = GalDelta(grid, sigma, z_mean, 'dummy', template_zmean_shifts, template_width_shifts, get_delta_p=False)
        phi_fid_array = g_d.phi_fid_array
        phi_perturbed_array = g_d.phi_perturbed_array
        self.chi_mean_fid = g_d.chi_mean_fid
        self.chi_sigma_fid = g_d.chi_sigma_fid

        # Calculate the galaxy lensing kernel
        # But first, a dummy run to pre-compile with numba
        utils.lens_efficiency_kernel(grid.chi_array, grid.chi_max_int, np.zeros_like(grid.chi_array))

        # Now in earnest
        g_fid = 3 / 2. * Planck18.Om0 * Planck18.H0.value ** 2 / c.value ** 2 * grid.chi_array * (
                1 + grid.z_array[0, :]) * utils.lens_efficiency_kernel(grid.chi_array, grid.chi_max_int, phi_fid_array)
        g_pert = 3 / 2. * Planck18.Om0 * Planck18.H0.value ** 2 / c.value ** 2 * grid.chi_array * (
                1 + grid.z_array[0, :]) * utils.lens_efficiency_kernel(grid.chi_array, grid.chi_max_int, phi_perturbed_array)

        # Go on to extract the alms at each chi, and so on
        super().__init__(grid, g_pert, g_fid, self.mask)


import numpy as np
import camb
from camb import model
from astropy.cosmology import Planck18
from scipy.interpolate import interp1d
try:
    # A lot of this is copied from Nick Kokron's anzu repository
    import pyccl as ccl
    from anzu.emu_funcs import LPTEmulator
    from velocileptors.LPT.cleft_fftw import CLEFT
    from velocileptors.EPT.cleft_kexpanded_resummed_fftw import RKECLEFT
    def compute_velocileptors_spectra(cosmovec, snapscale, use_physical_densities=True,
                                      use_sigma_8=True, kecleft=True, cleftobj=None):
        '''
        Returns a spline object which computes the cleft component spectra. Computed either in
        "full" CLEFT or in "k-expanded" CLEFT which allows for faster redshift dependence.
        Args:
            cosmovec : array-like
                Vector containing cosmology in the order (ombh2, omch2, w0, ns, sigma8, H0, Neff).
                If self.use_sigma_8 != True, then ln(A_s/10^{-10}) should be provided instead of sigma8.
            snapscale : float
                scale factor
            kecleft: bool
                Bool to check if the calculation is being made with
        Returns:
            cleft_aem : InterpolatedUnivariateSpline
                Spline that computes basis spectra as a function of k
        '''

        if use_physical_densities:
            if use_sigma_8:
                cosmo = ccl.Cosmology(Omega_b=cosmovec[0] / (cosmovec[5] / 100) ** 2,
                                      Omega_c=cosmovec[1] /
                                              (cosmovec[5] / 100) ** 2,
                                      h=cosmovec[5] / 100, n_s=cosmovec[3],
                                      w0=cosmovec[2], Neff=cosmovec[6],
                                      sigma8=cosmovec[4])
            else:
                cosmo = ccl.Cosmology(Omega_b=cosmovec[0] / (cosmovec[5] / 100) ** 2,
                                      Omega_c=cosmovec[1] /
                                              (cosmovec[5] / 100) ** 2,
                                      h=cosmovec[5] / 100, n_s=cosmovec[3],
                                      w0=cosmovec[2], Neff=cosmovec[6],
                                      A_s=np.exp(cosmovec[4]) * 1e-10)
        else:
            if use_sigma_8:
                cosmo = ccl.Cosmology(Omega_b=cosmovec[0],
                                      Omega_c=cosmovec[1] - cosmovec[0],
                                      h=cosmovec[5] / 100, n_s=cosmovec[3],
                                      w0=cosmovec[2], Neff=cosmovec[6],
                                      sigma8=cosmovec[4])
            else:
                cosmo = ccl.Cosmology(Omega_b=cosmovec[0],
                                      Omega_c=cosmovec[1] - cosmovec[0],
                                      h=cosmovec[5] / 100, n_s=cosmovec[3],
                                      w0=cosmovec[2], Neff=cosmovec[6],
                                      A_s=np.exp(cosmovec[4]) * 1e-10)

        k = np.logspace(-3, 1, 1000)

        if kecleft:
            # If using kecleft, check that we're only varying the redshift

            if cleftobj is None:
                # Do the full calculation again, as the cosmology changed.
                pk = ccl.linear_matter_power(
                    cosmo, k * cosmo['h'], 1) * (cosmo['h']) ** 3

                # Function to obtain the no-wiggle spectrum.
                # Not implemented yet, maybe Wallisch maybe B-Splines?
                # pnw = p_nwify(pk)
                # For now just use Stephen's standard savgol implementation.
                cleftobj = RKECLEFT(k, pk)

            # Adjust growth factors
            D = ccl.background.growth_factor(cosmo, snapscale)
            cleftobj.make_ptable(D=D, kmin=k[0], kmax=k[-1], nk=1000)
            cleftpk = cleftobj.pktable.T

        else:
            # Using "full" CLEFT, have to always do calculation from scratch
            pk = ccl.linear_matter_power(
                cosmo, k * cosmo['h'], snapscale) * (cosmo['h']) ** 3
            cleftobj = CLEFT(k, pk, N=2700, jn=10, cutoff=1)
            cleftobj.make_ptable()

            cleftpk = cleftobj.pktable.T

            # Different cutoff for other spectra, because otherwise different
            # large scale asymptote

            cleftobj = CLEFT(k, pk, N=2700, jn=5, cutoff=10)
            cleftobj.make_ptable()

        cleftpk[3:, :] = cleftobj.pktable.T[3:, :]
        cleftpk[2, :] /= 2
        cleftpk[6, :] /= 0.25
        cleftpk[7, :] /= 2
        cleftpk[8, :] /= 2

        cleftspline = interp1d(cleftpk[0], cleftpk, fill_value='extrapolate')

        return cleftspline, cleftobj

    def get_galaxy_ps_anzu(bvec, k, zs_sampled, halomatter=False):
        '''
        Calculate the galaxy power spectrum in the Planck 18 cosmology
        - Inputs:
            * bvec = list containing [b1,    b2,    bs2,   bnabla2, SN] to be fed to Anzu to obtain Pgg
            * z_mean = float. Central redshift of the fiducial dndz
            * k = np array of floats. k at which to evaluate Pkgg.
            * zs_sampled = redshifts at which to evaluate the Anzu prediction
            * halomatter (optional) = Bool. If False, get gg spectrum. If False, get galaxy-matter cross spectrum
        '''
        emu = LPTEmulator()
        h = Planck18.H0.value / 100.

        for i, z in enumerate(zs_sampled):
            a = 1 / (1 + z)
            if i == 0:
                cosmo_vec = np.atleast_2d([Planck18.Ob0 * h ** 2, Planck18.Odm0 * h ** 2, -1, 0.966, 0.812,
                                           Planck18.H0.value, 3.046, a])  # Values from Planck 2018
            else:
                cosmo_vec = np.vstack([np.atleast_2d([Planck18.Ob0 * h ** 2, Planck18.Odm0 * h ** 2, -1, 0.966, 0.812,
                                                      Planck18.H0.value, 3.046, a]), cosmo_vec])

        lpt_spec = np.zeros((len(cosmo_vec), 10, 700))

        # Evaluate predictions at the relevant redshifts
        for i, cv in enumerate(cosmo_vec):
            lpt_interp, cleftobk = compute_velocileptors_spectra(cv, cv[-1],
                                                                 use_physical_densities=emu.use_physical_densities,
                                                                 use_sigma_8=emu.use_sigma_8, kecleft=False)
            lpt_spec[i, ...] = lpt_interp(emu.k)[1:11, :]
        emu_spec = emu.predict(k, cosmo_vec, spec_lpt=lpt_spec)
        Pk = np.zeros((len(k), len(cosmo_vec[:, -1])))
        if halomatter:
            min_idx = len(k)
        else:
            min_idx = 0
        for i, z in enumerate(cosmo_vec[:, -1]):
            Pk[:, i] = emu.basis_to_full(k, bvec, emu_spec[i, :, :], halomatter=halomatter)[min_idx:]
        return Pk
except ImportError:
    print('Anzu/velocileptors/ccl not installed. Proceeding just with CAMB matter PS x linear galaxy bias')

def get_galaxy_ps(g_bias, zs_sampled, g2_bias=None, gbias_mode='linear'):
    '''
    Calculate the galaxy power spectrum
    - Inputs:
        * g_bias = galaxy bias. if gbias_mode=='anzu', a list containing Lagrangian bias [b1,    b2,    bs2,   bnabla2, SN],
                              if gbias_mode=='linear', a float with linear bias value at center of dndz
        * z_mean = float. Central redshift of the fiducial dndz
        * zs_sampled = redshifts at which to evaluate the prediction
        * g2_bias (optional) = Like g_bias, but for the second galaxy sample in spectrum. If None, get galaxy-matter cross-spectrum.
        * gbias_mode (optional) = 'linear' or 'anzu'. Galaxy bias prescription
    '''
    if g2_bias is None:
        halomatter = True
    # ToDo: Choose k's more systematically
    k = np.logspace(-3, 0, 200)
    if gbias_mode=='anzu':
        try:
            # TODO: implement different galaxy bias for two samples in anzu galaxy cross-spectrum
            Pk = get_galaxy_ps_anzu(g_bias, k, zs_sampled, halomatter=halomatter)
            return k, Pk
        except:
            print('Anzu/velocileptors not installed. Proceeding just with CAMB matter PS x linear bias')
    k, pk_nonlin = get_matter_ps(zs_sampled)
    Pk = np.swapaxes(pk_nonlin, 0, 1)
    try:
        if g2_bias is None:
            # Halo-matter cross-spectrum using linear galaxy bias
            return k, Pk * g_bias
        else:
            # Galaxy auto-spectrum using linear galaxy bias
            return k, Pk * g_bias * g2_bias
    except:
        print('Galaxy bias must be a linear (i.e. a single number) when not using anzu')

def get_matter_ps(redshifts):
    #Now get matter power spectra and sigma8 at redshifts between 0 and sufficiently behind the perturbed sources
    pars = camb.CAMBparams()
    h = Planck18.H0.value/100.
    pars.set_cosmology(H0=Planck18.H0.value, ombh2=Planck18.Ob0 * h**2, omch2=Planck18.Odm0 * h**2)
    pars.InitPower.set_params(ns=0.966)
    #Note non-linear corrections couples to smaller scales than you want
    pars.set_matter_power(redshifts=redshifts, kmax=2.0)

    #Linear spectra
    pars.NonLinear = model.NonLinear_none
    results = camb.get_results(pars)
    kh, z, pk = results.get_matter_power_spectrum(minkh=1e-4, maxkh=1e2, npoints = 500)
    s8 = np.array(results.get_sigma8())

    #Non-Linear spectra (Halofit)
    pars.NonLinear = model.NonLinear_both
    results.calc_power_spectra(pars)
    kh_nonlin, z_nonlin, pk_nonlin = results.get_matter_power_spectrum(minkh=1e-4, maxkh=1e2, npoints = 500)

    # Remove factors of h
    k_nonlin = kh_nonlin * h
    pk_nonlin /= h**3
    return k_nonlin, pk_nonlin



import numpy as np
import healpy as hp
import numba as nb

def find_closest_indices(chi_array, ref_array):
    overlapping = np.logical_and(ref_array > np.min(chi_array), ref_array < np.max(chi_array))
    out = np.zeros(np.sum(overlapping))
    for i, a in enumerate(ref_array[overlapping]):
        out[i] = np.abs(chi_array - a).argmin()
    return out.astype(int)

def bl(fwhm_arcmin, lmax):
    """ returns the map-level transfer function for a symmetric Gaussian beam.
         * fwhm_arcmin      = beam full-width-at-half-maximum (fwhm) in arcmin.
         * lmax             = maximum multipole.
    """
    ls = np.arange(0, lmax + 1)
    return np.exp(-(fwhm_arcmin * np.pi / 180. / 60.) ** 2 / (16. * np.log(2.)) * ls * (ls + 1.))

def generate_template(nside, alpha, target_std, lmax_pert):
    '''
    Generate a pixelized map where each pixel value corresponds to a random shift of the
    center of the dndz. The map has a standard deviation set by target_std, and some underlying
    power-law power spectrum for the shifts
    Inputs:
        * nside = int. Defines the pixelization
        * alpha = float. Power law index for the perturbation Cls
        * target_std = float. Standard deviation of the resulting map
        * lmax_pert = int. lmax of the Cls
    '''
    pert_cls = np.arange(lmax_pert, dtype=float)**alpha
    pert_cls[0]=0

    # Convert to a map
    sim_map = hp.synfast(pert_cls, nside)

    # Normalize Cls to give desired variance
    norm_factor_for_alms = np.nan_to_num(target_std**2 / np.var(sim_map, ddof=1))**0.5
    return norm_factor_for_alms*sim_map

@nb.jit(nopython=True)
def integral_at_chi(chi, chi_max_int, phi_array, chi_s):
    min_idx = np.where(chi_s>=chi)[0][0]

    if phi_array.ndim==2:
        outs_at_chi = np.zeros(phi_array.shape[0], dtype=float)
        for j in range(phi_array.shape[0]):
            # Iterate over healpixels
            integrand = phi_array[j,:] * (chi_s - chi)/chi_s
            outs_at_chi[j] = np.trapz(integrand[min_idx:], chi_s[min_idx:])
    else:
        integrand = phi_array * (chi_s - chi)/chi_s
        outs_at_chi = np.trapz(integrand[min_idx:], chi_s[min_idx:])
    return outs_at_chi

@nb.jit(nopython=True)
def lens_efficiency_kernel(chis, chi_max_int, phi_array):
    output = np.zeros_like(phi_array)
    if phi_array.ndim==1:
        for i, chi in enumerate(chis):
            output[i] = integral_at_chi(chi, chi_max_int, phi_array, chis)
    else:
        for i, chi in enumerate(chis):
            output[:, i] = integral_at_chi(chi, chi_max_int, phi_array, chis)
    return output